/* signal.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#include "signal.h"
#include "bots.h"
#include "ChanNode.h"
#include "ClientSocket.h"
#include "IOHandler.h"
#include "ConfigParser.h"
#include "log.h"

static void sigcrash();
static void sigexit();

void sighandler(int signum) {
    printf_log("main", LOG_INFO, "Received Signal %d\n", signum);
    signal(signum, SIG_DFL);
    switch(signum) {
        case SIGABRT:
        case SIGTERM:
        case SIGINT:
            //normal shutdown
            sigexit(signum);
            break;
        case SIGSEGV:
        case SIGFPE:
        case SIGILL:
            //crash
            sigcrash(signum);
            break;
    }
    #ifdef WIN32
    exit(signum);
    #else
    kill(getpid(), signum);
    #endif
}

static void sigcrash(int signum) {
    char coregen[MAXLEN];
	coregen[0] = 0;
    char *signame;
    switch(signum) {
        case SIGSEGV:
            signame = "SIGSEGV";
            break;
        case SIGFPE:
            signame = "SIGFPE";
            break;
        case SIGILL:
            signame = "SIGILL";
            break;
        default:
            signame = "SIGUNKNOWN";
            break;
    }
    printf_log("main", LOG_ERROR, "NeonServ process crashed (%s)\n", signame);
    #ifndef WIN32
    char gcore[50];
    sprintf(gcore, "gcore %u", getpid());
    int sysretn = system(gcore); //generate core file
    sprintf(coregen, "core file generated. (%d)", sysretn);
    printf_log("main", LOG_ERROR | LOG_INFO, "generated core file.\n");
    #endif
    //close all bots
    struct ClientSocket *bot;
    for(bot = getBots(SOCKET_FLAG_READY, NULL); bot; bot = getBots(SOCKET_FLAG_READY, bot)) {
        if((bot->flags & SOCKET_FLAG_CONNECTED)) {
            iohandler_close(bot->iofd);
            bot->flags &= ~(SOCKET_FLAG_CONNECTED | SOCKET_FLAG_READY);
            bot->iofd = NULL;
        }
    }
    printf_log("main", LOG_INFO, "hard shutdown...\n");
    usleep(2000000);
    //hard restart
    restart_bot(1);
    exit(0);
}

static void sigexit(int signum) {
    stop_bot();
}
